/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2019, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 /*
  *
  * Implementation of template list that keeps data items
  *
  * $Created by: HU Chi (huchinlp@foxmail.com)
  *
  */

#include "XMem.h"
#include "XGlobal.h"
#ifndef __XLIST_H__
#define __XLIST_H__


/* the nts (NiuTrans.Tensor) namespace */
namespace nts {

/* the XListType class */
template <typename T>
class XListType {
public:
    typedef int (*ListCompare)(const T* item1, const T* item2);

    /* data items */
    T** items;

    /* number of items */
    int count;

    /* maximum number of items can be kept */
    int maxNum;

    /* the memory pool for data array allocation */
    XMem* mem;

    /* indicates whether data items are integers */
    bool isIntList;

public:
    /* constructor */
    XListType();

    /* constructor */
    XListType(int myMaxNum, bool isIntListOrNot = false);

    /* constructor */
    XListType(int myMaxNum, XMem* myMem, bool isIntListOrNot = false);

    /* de-constructor */
    ~XListType();

    /* add an item into the list */
    void Add(T* item);

	/* add a number of items into the list */
    void Add(T** inputItems, int inputItemCount);

	/* append a list to the current list */
    void AddList(XListType* l);

	/* insert an item to the given position of the list */
    void Insert(int pos, T* item);

	/* get the item at position i */
    T* GetItem(int i) const;

	/* set the item at position i */
    void SetItem(int i, T* item);

	/* find the position of the first matched item  */
    int FindFirst(T* item);

	/* clear the data array */
    void Clear();

	/* sort the list */
    void Sort(int itemSize, ListCompare comp);

	/* reverse the list */
    void Reverse();

	/* remove the item at position i */
    void Remove(int i);

	/* copy the list */
    XListType* Copy(XMem* myMem);

	/* shuffle the list */
    void Shuffle(int nround = 10, int beg = -1, int len = 0);

    /* short */
    T* Get(int i) { return GetItem(i); };
	void Set(int i, T* item) { SetItem(i, item); };
};

/* constructor */
template <typename T>
XListType<T>::XListType()
{
    mem = NULL;
    maxNum = 0;
    count = 0;
    items = NULL;
    isIntList = false;
}

/* 
constructor 
>> myMaxNum - maximum number of items to keep
>> isIntListOrNot - specify if the list keeps int items
*/
template <typename T>
XListType<T>::XListType(int myMaxNum, bool isIntListOrNot)
{
    mem = NULL;
    maxNum = myMaxNum;
    count = 0;
    items = new T*[myMaxNum];
    isIntList = isIntListOrNot;
}

/* 
constructor 
>> myMaxNum - maximum number of items to keep
>> myMem - the memory pool used for data allocation
>> isIntListOrNot - specify if the list keeps int items
*/
template <typename T>
XListType<T>::XListType(int myMaxNum, XMem* myMem, bool isIntListOrNot)
{
    mem = myMem;
    maxNum = myMaxNum;
    count = 0;
    items = (T**)mem->Alloc(mem->devID, sizeof(T*) * maxNum);
    isIntList = isIntListOrNot;
}

/* de-constructor */
template <typename T>
XListType<T>::~XListType()
{
    if (isIntList) {
        for (int i = 0; i < count; i++) {
            int* p = (int*)items[i];
            delete[] p;
        }
    }
    if (mem == NULL)
        delete[] items;
}


/*
add an item into the list
>> item - pointer to the item
*/
template <typename T>
void XListType<T>::Add(T* item)
{
    if (count == maxNum) {
        T** newItems;
        if (mem == NULL)
            newItems = new T*[maxNum * 2 + 1];
        else
            newItems = (T**)mem->Alloc(mem->devID, sizeof(T*) * (maxNum * 2 + 1));
        memcpy(newItems, items, sizeof(T*) * maxNum);
        if (mem == NULL)
            delete[] items;
        items = newItems;
        maxNum = maxNum * 2 + 1;
    }

    items[count++] = item;
}

/* 
add a number of items into the list 
>> inputItems - pointer to the array of items
>> inputItemCount - number of input items
*/
template <typename T>
void XListType<T>::Add(T** inputItems, int inputItemCount)
{
    if (count + inputItemCount >= maxNum) {
        int newMaxNum = (count + inputItemCount) * 2 + 1;
        T** newItems;
        if (mem == NULL)
            newItems = new T*[newMaxNum];
        else
            newItems = (T**)mem->Alloc(mem->devID, sizeof(T*) * newMaxNum);
        memcpy(newItems, items, sizeof(T*) * maxNum);
        if (mem == NULL)
            delete[] items;
        items = newItems;
        maxNum = newMaxNum;
    }
    memcpy(items + count, inputItems, sizeof(T*) * inputItemCount);
    count += inputItemCount;
}

/*
append a list to the current list
>> l - the list we use to append
*/
template <typename T>
void XListType<T>::AddList(XListType* l)
{
    Add(l->items, l->count);
}

/*
insert an item to the given position of the list
>> pos - the position
>> item - the item for insertion
*/
template <typename T>
void XListType<T>::Insert(int pos, T* item)
{
    if (count == maxNum) {
        T** newItems;
        if (mem == NULL)
            newItems = new T*[maxNum * 2 + 1];
        else
            newItems = (T**)mem->Alloc(mem->devID, sizeof(T*) * (maxNum * 2 + 1));
        memcpy(newItems, items, sizeof(T*) * maxNum);
        if (mem == NULL)
            delete[] items;
        items = newItems;
        maxNum = maxNum * 2 + 1;
    }

    for (int i = count - 1; i >= pos; i--)
        items[i + 1] = items[i];
    items[pos] = item;
    count++;
}

/* get the item at position i */
template <typename T>
T* XListType<T>::GetItem(int i) const
{
    CheckNTErrors(i >= -1 && i < count, "Index of a list item is out of scope!");
    CheckNTErrors(count > 0, "Cannt index the item in an empty list!");
    if (i == -1)
        return items[count - 1];
    else
        return items[i];
}

/* set the item at position i */
template <typename T>
void XListType<T>::SetItem(int i, T* item)
{
    if (i >= 0 && i < count)
        items[i] = item;
}

/* 
find the position of the first matched item 
>> item - the item for matching
<< the position where we hit the item (if any)
*/

template <typename T>
int XListType<T>::FindFirst(T* item)
{
    for (int i = 0; i < count; i++) {
        if (item == items[i])
            return i;
    }
    return -1;
}

/* clear the data array */
template <typename T>
void XListType<T>::Clear()
{
    if (isIntList) {
        for (int i = 0; i < count; i++) {
            delete[](int*) items[i];
        }
        count = 0;
    } else
        count = 0;
}

/* 
sort the list 
>> itemSize - size of an item
>> comp - the comparison function used in sorting
*/
template <typename T>
void XListType<T>::Sort(int itemSize, ListCompare comp)
{
    qsort(items, count, itemSize, comp);
}

/* reverse the list */
template <typename T>
void XListType<T>::Reverse()
{
    int half = count / 2;
    for (int i = 0; i < half; i++) {
        T* tmp = items[i];
        items[i] = items[count - i - 1];
        items[count - i - 1] = tmp;
    }
}

/* remove the item at position i */
template <typename T>
void XListType<T>::Remove(int i)
{
    if (i >= count || i < 0)
        return;

    memcpy(items + i, items + i + 1, sizeof(T*) * (count - i - 1));

    count--;
}

/* 
copy the list 
>> myMem - memory pool used for allocating the data in the new list
<< hard copy of the list
*/
template <typename T>
XListType<T>* XListType<T>::Copy(XMem* myMem)
{
    XListType<T>* newList = new XListType<T>(maxNum, myMem);
    for (int i = 0; i < count; i++) {
        newList->Add(GetItem(i));
    }
    return newList;
}

/* 
shuffle the list
>> nround - number of rounds for shuffling
>> beg - where we start
>> len - how many items are used in shuffling
*/
template <typename T>
void XListType<T>::Shuffle(int nround, int beg, int len)
{
    if (beg < 0) {
        beg = 0;
        len = count;
    }

    if (beg + len > count)
        return;

    srand((unsigned int)time(NULL));

    for (int k = 0; k < nround; k++) {
        /* FisherCYates shuffle */
        for (int i = 0; i < len; i++) {
            float a = (float)rand() / RAND_MAX;
            size_t j = (unsigned int)(a * (i + 1));
            T* t = items[beg + j];
            items[beg + j] = items[beg + i];
            items[beg + i] = t;
        }
    }
}

struct XTensor;

/* typedef for list */
typedef XListType<int> IntList;
typedef XListType<char> CharList;
typedef XListType<long> LongList;
typedef XListType<float> FloatList;
typedef XListType<short> ShortList;
typedef XListType<XTensor> XList;

} /* end of the nts (NiuTrans.Tensor) namespace */

#endif // __XLIST_H__
